// ------- BEGIN MA5 CHANGE -------

/** \class Isolation
 *
 *  Sums transverse momenta of isolation objects (tracks, calorimeter towers, etc)
 *  within a DeltaR cone around a candidate and calculates fraction of this sum
 *  to the candidate's transverse momentum. outputs candidates that have
 *  the transverse momenta fraction within (PTRatioMin, PTRatioMax].
 *
 *  $Date: 2013-08-19 15:40:54 +0200 (Mon, 19 Aug 2013) $
 *  $Revision: 1267 $
 *
 *
 *  \author P. Demin - UCL, Louvain-la-Neuve
 *
 */

#include "modules/IsolationCalculation.h"

#include "classes/DelphesClasses.h"
#include "classes/DelphesFactory.h"
#include "classes/DelphesFormula.h"

#include "ExRootAnalysis/ExRootResult.h"
#include "ExRootAnalysis/ExRootFilter.h"
#include "ExRootAnalysis/ExRootClassifier.h"

#include "TMath.h"
#include "TString.h"
#include "TFormula.h"
#include "TRandom3.h"
#include "TObjArray.h"
#include "TDatabasePDG.h"
#include "TLorentzVector.h"

#include <algorithm>
#include <stdexcept>
#include <iostream>
#include <sstream>
#include <fstream>

using namespace std;

//------------------------------------------------------------------------------

class IsolationCalculationClassifier : public ExRootClassifier
{
public:

  IsolationCalculationClassifier() {}

  Int_t GetCategory(TObject *object);

  Double_t fPTMin;
};

//------------------------------------------------------------------------------

Int_t IsolationCalculationClassifier::GetCategory(TObject *object)
{
  Candidate *track = static_cast<Candidate*>(object);
  const TLorentzVector &momentum = track->Momentum;

  if(momentum.Pt() < fPTMin) return -1;

  return 0;
}


IsolationCalculation::IsolationCalculation() :
  fClassifier(0), fFilter(0), fOutputArray(0),
  fItTrackInputArray(0), fItCaloTowerInputArray(0), fItCandidateInputArray(0)
{
  fClassifier = new IsolationCalculationClassifier;
}

//------------------------------------------------------------------------------

IsolationCalculation::~IsolationCalculation()
{
}

//------------------------------------------------------------------------------

void IsolationCalculation::Init()
{
  // Import parameters
  fClassifier->fPTMin = GetDouble("PTMin", 0.5);

  // Create output arrays
  fOutputArray = ExportArray(GetString("OutputArray", "DelfesElectrons"));

  // Import input arrays
  fTrackInputArray     = ImportArray(GetString("TrackInputArray",     "Delphes/partons"));
  fCaloTowerInputArray = ImportArray(GetString("CaloTowerInputArray", "Delphes/partons"));
  fCandidateInputArray = ImportArray(GetString("CandidateInputArray", "Calorimeter/electrons"));

  // Iterator
  fItTrackInputArray     = fTrackInputArray->MakeIterator();
  fItCaloTowerInputArray = fCaloTowerInputArray->MakeIterator();
  fItCandidateInputArray = fCandidateInputArray->MakeIterator();

  // Filter on arrays
  fFilter = new ExRootFilter(fTrackInputArray);

  // PileUp
  const char *rhoInputArrayName;
  rhoInputArrayName = GetString("RhoInputArray", "");
  if(rhoInputArrayName[0] != '\0') fRhoInputArray = ImportArray(rhoInputArrayName);
  else fRhoInputArray = 0;
}

//------------------------------------------------------------------------------

void IsolationCalculation::Finish()
{
  if (fFilter) delete fFilter;
  if (fItCandidateInputArray) delete fItCandidateInputArray;
  if (fItCaloTowerInputArray) delete fItCaloTowerInputArray;
  if (fItTrackInputArray)     delete fItTrackInputArray;
}

//------------------------------------------------------------------------------

void IsolationCalculation::Process()
{
  Candidate *candidate, *isolation;
  TObjArray *trackArray;

  // Pile-Up
  Double_t rho = 0.0;
  if (fRhoInputArray && fRhoInputArray->GetEntriesFast() > 0)
  {
    candidate = static_cast<Candidate*>(fRhoInputArray->At(0));
    rho = candidate->Momentum.Pt();
  }

  // Select tracks for isolation
  fFilter->Reset();
  trackArray = fFilter->GetSubArray(fClassifier, 0);
  if(trackArray == 0) return;
  TIter itTrackArray(trackArray);

  // loop over all input jets
  fItCandidateInputArray->Reset();
  while((candidate = static_cast<Candidate*>(fItCandidateInputArray->Next())))
  {
    const TLorentzVector &candidateMomentum = candidate->Momentum;

    // loop over all input tracks
    Double_t sumPT05   = 0.0;
    Double_t sumET05   = 0.0;
    Int_t    counter05 = 0;
    Double_t sumPT04   = 0.0;
    Double_t sumET04   = 0.0;
    Int_t    counter04 = 0;
    Double_t sumPT03   = 0.0;
    Double_t sumET03   = 0.0;
    Int_t    counter03 = 0;
    Double_t sumPT02   = 0.0;
    Double_t sumET02   = 0.0;
    Int_t    counter02 = 0;

    itTrackArray.Reset();
    while((isolation = static_cast<Candidate*>(itTrackArray.Next())))
    {
      const TLorentzVector &isolationMomentum = isolation->Momentum;
      double DR = candidateMomentum.DeltaR(isolationMomentum);
      
      // overlap
      if (candidate->Overlaps(isolation)) continue;

      // Cone with DeltaR=0.5
      if(DR<=0.5)
      {
        sumPT05 += isolationMomentum.Pt();
        ++counter05;
      }

      // Cone with DeltaR=0.4
      if(DR<=0.4)
      {
        sumPT04 += isolationMomentum.Pt();
        ++counter04;
      }

      // Cone with DeltaR=0.3
      if(DR<=0.3)
      {
        sumPT03 += isolationMomentum.Pt();
        ++counter03;
      }

      // Cone with DeltaR=0.2
      if(DR<=0.2)
      {
        sumPT02 += isolationMomentum.Pt();
        ++counter02;
      }

    }

    fItCaloTowerInputArray->Reset();
    while((isolation = static_cast<Candidate*>(fItCaloTowerInputArray->Next())))
    {
      const TLorentzVector &isolationMomentum = isolation->Momentum;
      double DR = candidateMomentum.DeltaR(isolationMomentum);
      
      // Cone with DeltaR=0.5
      if(DR<=0.5)
      {
        sumET05 += isolationMomentum.Pt();
      }

      // Cone with DeltaR=0.4
      if(DR<=0.4)
      {
        sumET04 += isolationMomentum.Pt();
      }

      // Cone with DeltaR=0.3
      if(DR<=0.3)
      {
        sumET03 += isolationMomentum.Pt();
      }

      // Cone with DeltaR=0.2
      if(DR<=0.2)
      {
        sumET02 += isolationMomentum.Pt();
      }

    }

    // Correct sum for pile-up contamination
    sumPT05 -= rho*0.5*0.5*TMath::Pi();  
    sumPT04 -= rho*0.4*0.4*TMath::Pi();  
    sumPT03 -= rho*0.3*0.3*TMath::Pi();  
    sumPT02 -= rho*0.2*0.2*TMath::Pi();  

    // Clone the candidate 
    candidate = static_cast<Candidate*>(candidate->Clone());
    candidate->sumPT05 = sumPT05;
    candidate->sumPT04 = sumPT04;
    candidate->sumPT03 = sumPT03;
    candidate->sumPT02 = sumPT02;
    candidate->sumET05 = sumET05;
    candidate->sumET04 = sumET04;
    candidate->sumET03 = sumET03;
    candidate->sumET02 = sumET02;
    candidate->nTrack05 = counter05;
    candidate->nTrack04 = counter04;
    candidate->nTrack03 = counter03;
    candidate->nTrack02 = counter02;

    // Save the new candidate
    fOutputArray->Add(candidate);
  }
}

//------------------------------------------------------------------------------

// ------- END MA5 CHANGE -------
